"""
Utility methods to print system info for debugging

adapted from ``pandas.show_versions``
adapted from ``sklearn.show_versions``
"""
# License: BSD 3 clause

import platform
import sys

_pmdarima_deps = (
    # setuptools needs to be before pip: https://github.com/pypa/setuptools/issues/3044#issuecomment-1024972548 # noqa:E501
    "setuptools",
    "pip",
    "sklearn",
    "statsmodels",
    "numpy",
    "scipy",
    "Cython",
    "pandas",
    "joblib",
    "pmdarima",
)

# Packages that have a different import name than name on PyPI
_install_mapping = {
    "sklearn": "scikit-learn"
}


def _get_sys_info():
    """System information

    Return
    ------
    sys_info : dict
        system and Python version information

    """
    python = sys.version.replace('\n', ' ')

    blob = [
        ("python", python),
        ('executable', sys.executable),
        ("machine", platform.platform()),
    ]

    return dict(blob)


def _get_deps_info(deps=_pmdarima_deps):
    """Overview of the installed version of main dependencies

    Returns
    -------
    deps_info: dict
        version information on relevant Python libraries
    """
    def get_version(module):
        return module.__version__

    deps_info = {}

    # TODO: We can get rid of this when we deprecate 3.7
    if sys.version_info.minor <= 7:
        import importlib

        for modname in deps:
            try:
                if modname in sys.modules:
                    mod = sys.modules[modname]
                else:
                    mod = importlib.import_module(modname)
                ver = get_version(mod)
                deps_info[modname] = ver
            except ImportError:
                deps_info[modname] = None

    else:
        from importlib.metadata import PackageNotFoundError, version

        for modname in deps:
            try:
                deps_info[modname] = version(_install_mapping.get(modname, modname))  # noqa:E501
            except PackageNotFoundError:
                deps_info[modname] = None

    return deps_info


def show_versions():
    """Print useful debugging information"""
    sys_info = _get_sys_info()
    deps_info = _get_deps_info()

    print('\nSystem:')
    for k, stat in sys_info.items():
        print("{k:>10}: {stat}".format(k=k, stat=stat))

    print('\nPython dependencies:')
    for k, stat in deps_info.items():
        print("{k:>11}: {stat}".format(k=k, stat=stat))



# -*- coding: utf-8 -*-
#
# Author: Taylor G Smith <taylor.smith@alkaline-ml.com>
#
# Plotting wrapper functions

from .array import check_endog
from ..compat.pandas import plotting as pd_plotting
from ..compat.matplotlib import get_compatible_pyplot

from statsmodels.graphics import tsaplots

import numpy as np
import os

# User may not have matplotlib
try:
    # Gets the MPL.pyplot import (combatibilitized). Only use debug mode if set
    # on the machine in an environment variable
    debug = os.environ.get("PMDARIMA_MPL_DEBUG", "false").lower() == "true"

    # If it's a Travis CI machine, we want to set the backend via env variable
    backend = os.environ.get("PMD_MPL_BACKEND", None)
    mpl = get_compatible_pyplot(backend=backend, debug=debug)

except ImportError:
    mpl = None

__all__ = [
    'autocorr_plot',
    'decomposed_plot',
    'plot_acf',
    'plot_pacf',
    'tsdisplay',
]


def _err_for_no_mpl():
    if mpl is None:
        # Per Issue #47:
        raise ImportError(
            "You do not have matplotlib installed. In order to "
            "create plots, you'll need to pip install matplotlib!")


def _get_plt():
    """Get MPL pyplot if it exists or raise an error if not"""
    _err_for_no_mpl()
    return mpl


def _show_or_return(obj, show):
    if show:
        # We never cover this in tests, unfortunately. Even with the
        # cleanup tag, Travis doesn't play super nice with showing and
        # closing lots of plots over and over. But it's just one line...
        mpl.show()
        # returns None implicitly
    else:
        return obj


def decomposed_plot(decomposed_tuple, figure_kwargs=None, show=True):
    """Plot the decomposition of a time series.

    Plots the results of the time series decomposition in four plots:
    the 'x', 'trend', 'seasonal', and 'random' components.

    Parameters
    ----------
    decomposed_tuple : tuple, namedtuple or iterable
        Named tuple of series that consist of data, trend, seasonal, and
        random. Should be the result of :func:`pmdarima.arima.decompose`.

    figure_kwargs : dict, optional (default=None)
        Optional dictionary of keyword arguments that are passed to figure.

    show : bool, optional (default=True)
        Whether to show the plot after it's been created. If not, will return
        the plot as an Axis object instead.

    Notes
    -----
    This method will only show the plot if ``show=True`` (which is the default
    behavior). To simply get the axis back (say, to add to another canvas),
    use ``show=False``.
    """

    _err_for_no_mpl()

    fig, axes = mpl.subplots(4, 1, sharex=True, **figure_kwargs)

    y_labels = ['data', 'trend', 'seasonal', 'random']

    for ax in axes.flat:
        ax.set(ylabel=y_labels.pop(0))

    # Unpack positionally so that user could pass as a tuple, if they have
    # pre-computed...
    x, trend, ssnl, rand = decomposed_tuple

    axes[0].plot(x)
    axes[1].plot(trend)
    axes[2].plot(ssnl)
    axes[3].plot(rand)

    return _show_or_return(axes, show)


def autocorr_plot(series, show=True):
    """Plot a series' auto-correlation.

    A wrapper method for the Pandas ``autocorrelation_plot`` method.

    Parameters
    ----------
    series : array-like, shape=(n_samples,)
        The series or numpy array for which to plot an auto-correlation.

    show : bool, optional (default=True)
        Whether to show the plot after it's been created. If not, will return
        the plot as an Axis object instead.

    Notes
    -----
    This method will only show the plot if ``show=True`` (which is the default
    behavior). To simply get the axis back (say, to add to another canvas),
    use ``show=False``.

    Examples
    --------
    >>> autocorr_plot([1, 2, 3], False)  # doctest: +SKIP
    <matplotlib.axes._subplots.AxesSubplot object at 0x127f41dd8>

    Returns
    -------
    res : Axis or None
        If ``show`` is True, does not return anything. If False, returns
        the Axis object.
    """
    _err_for_no_mpl()
    res = pd_plotting.autocorrelation_plot(series)
    return _show_or_return(res, show)


def plot_acf(series, ax=None, lags=None, alpha=None, use_vlines=True,
             unbiased=False, fft=True, title='Autocorrelation',
             zero=True, vlines_kwargs=None, show=True, **kwargs):
    """Plot a series' auto-correlation as a line plot.

    A wrapper method for the statsmodels ``plot_acf`` method.

    Parameters
    ----------
    series : array-like, shape=(n_samples,)
        The series or numpy array for which to plot an auto-correlation.

    ax : Matplotlib AxesSubplot instance, optional
        If given, this subplot is used to plot in instead of a new figure being
        created.

    lags : int, array-like or None, optional (default=None)
        int or Array of lag values, used on horizontal axis. Uses
        np.arange(lags) when lags is an int.  If not provided,
        ``lags=np.arange(len(corr))`` is used.

    alpha : scalar, optional (default=None)
        If a number is given, the confidence intervals for the given level are
        returned. For instance if alpha=.05, 95 % confidence intervals are
        returned where the standard deviation is computed according to
        Bartlett's formula. If None, no confidence intervals are plotted.

    use_vlines : bool, optional (default=True)
        If True, vertical lines and markers are plotted.
        If False, only markers are plotted.  The default marker is 'o'; it can
        be overridden with a ``marker`` kwarg.

    unbiased : bool, optional (default=False)
        If True, then denominators for autocovariance are n-k, otherwise n

    fft : bool, optional (default=True)
        If True, computes the ACF via FFT.

    title : str, optional (default='Autocorrelation')
        Title to place on plot. Default is 'Autocorrelation'

    zero : bool, optional (default=True)
        Flag indicating whether to include the 0-lag autocorrelation.
        Default is True.

    vlines_kwargs : dict, optional (default=None)
        Optional dictionary of keyword arguments that are passed to vlines.

    show : bool, optional (default=True)
        Whether to show the plot after it's been created. If not, will return
        the plot as an Axis object instead.

    **kwargs : kwargs, optional
        Optional keyword arguments that are directly passed on to the
        Matplotlib ``plot`` and ``axhline`` functions.

    Notes
    -----
    This method will only show the plot if ``show=True`` (which is the default
    behavior). To simply get the axis back (say, to add to another canvas),
    use ``show=False``.

    Examples
    --------
    >>> plot_acf([1, 2, 3], show=False)  # doctest: +SKIP
    <matplotlib.figure.Figure object at 0x122fab4e0>

    Returns
    -------
    plt : Axis or None
        If ``show`` is True, does not return anything. If False, returns
        the Axis object.
    """
    _err_for_no_mpl()
    res = tsaplots.plot_acf(
        x=series, ax=ax, lags=lags, alpha=alpha, use_vlines=use_vlines,
        unbiased=unbiased, fft=fft, title=title, zero=zero,
        vlines_kwargs=vlines_kwargs, **kwargs)

    return _show_or_return(res, show)


def plot_pacf(series, ax=None, lags=None, alpha=None, method='yw',
              use_vlines=True, title='Partial Autocorrelation', zero=True,
              vlines_kwargs=None, show=True, **kwargs):
    """Plot a series' partial auto-correlation as a line plot.

    A wrapper method for the statsmodels ``plot_pacf`` method.

    Parameters
    ----------
    series : array-like, shape=(n_samples,)
        The series or numpy array for which to plot an auto-correlation.

    ax : Matplotlib AxesSubplot instance, optional
        If given, this subplot is used to plot in instead of a new figure being
        created.

    lags : int, array-like or None, optional (default=None)
        int or Array of lag values, used on horizontal axis. Uses
        np.arange(lags) when lags is an int.  If not provided,
        ``lags=np.arange(len(corr))`` is used.

    alpha : scalar, optional (default=None)
        If a number is given, the confidence intervals for the given level are
        returned. For instance if alpha=.05, 95 % confidence intervals are
        returned where the standard deviation is computed according to
        Bartlett's formula. If None, no confidence intervals are plotted.

    method : str, optional (default='yw')
        Specifies which method for the calculations to use. One of
        {'ywunbiased', 'ywmle', 'ols', 'ld', 'ldb', 'ldunbiased', 'ldbiased'}:

        - yw or ywunbiased : yule walker with bias correction in denominator
          for acovf. Default.
        - ywm or ywmle : yule walker without bias correction
        - ols - regression of time series on lags of it and on constant
        - ld or ldunbiased : Levinson-Durbin recursion with bias correction
        - ldb or ldbiased : Levinson-Durbin recursion without bias correction

    use_vlines : bool, optional (default=True)
        If True, vertical lines and markers are plotted.
        If False, only markers are plotted.  The default marker is 'o'; it can
        be overridden with a ``marker`` kwarg.

    title : str, optional (default='Partial Autocorrelation')
        Title to place on plot. Default is 'Partial Autocorrelation'

    zero : bool, optional (default=True)
        Flag indicating whether to include the 0-lag autocorrelation.
        Default is True.

    vlines_kwargs : dict, optional (default=None)
        Optional dictionary of keyword arguments that are passed to vlines.

    show : bool, optional (default=True)
        Whether to show the plot after it's been created. If not, will return
        the plot as an Axis object instead.

    **kwargs : kwargs, optional
        Optional keyword arguments that are directly passed on to the
        Matplotlib ``plot`` and ``axhline`` functions.

    Notes
    -----
    This method will only show the plot if ``show=True`` (which is the default
    behavior). To simply get the axis back (say, to add to another canvas),
    use ``show=False``.

    Examples
    --------
    >>> plot_pacf([1, 2, 3, 4], show=False)  # doctest: +SKIP
    <matplotlib.figure.Figure object at 0x129df1630>

    Returns
    -------
    plt : Axis or None
        If ``show`` is True, does not return anything. If False, returns
        the Axis object.
    """
    _err_for_no_mpl()
    res = tsaplots.plot_pacf(
        x=series, ax=ax, lags=lags, alpha=alpha, method=method,
        use_vlines=use_vlines, title=title, zero=zero,
        vlines_kwargs=vlines_kwargs, **kwargs)

    return _show_or_return(res, show)


def tsdisplay(y, lag_max=50, figsize=(8, 6), title=None, bins=25,
              series_kwargs=None, acf_kwargs=None, hist_kwargs=None,
              show=True):
    """Display the time series and some of its key statistics

    The equivalent of R's ``forecast::tsdisplay``, showing the series, the
    histogram and the ACF plot.

    Parameters
    ----------
    y : array-like, shape=(n_samples,)
        The series or numpy array for which to plot an auto-correlation.

    lag_max : int, optional (default=50)
        The number of lags for the ACF plot

    figsize : tuple, optional (default=(8, 6))
        The size of the figure

    title : str, optional (default=None)
        A title for the series, if any.

    bins : int, optional (default=25)
        The number of bins for the histogram

    series_kwargs : dict or None, optional (default=None)
        Keyword arguments to pass when plotting the series

    acf_kwargs : dict or None, optional (default=None)
        Keyword arguments to pass when plotting the ACF

    hist_kwargs : dict or None, optional (default=None)
        Keyword arguments to pass when plotting the histogram

    show : bool, optional (default=True)
        Whether to show the plot after it's been created. If not, will return
        the plot as a Figure object instead.

    Examples
    --------
    >>> import pmdarima as pm
    >>> tsdisplay(pm.datasets.load_sunspots(), show=False)
    <Figure size 800x600 with 3 Axes>

    Returns
    -------
    plt : Figure or None
        If ``show`` is True, does not return anything. If False, returns
        the Figure object.
    """

    _err_for_no_mpl()
    from matplotlib import gridspec

    fig = mpl.figure(figsize=figsize)
    gs = gridspec.GridSpec(4, 2)
    ax0 = fig.add_subplot(gs[0:2, 0:])
    ax1 = fig.add_subplot(gs[2:, 0])
    ax2 = fig.add_subplot(gs[2:, 1])

    # make sure y is a np array
    y = check_endog(y, copy=False, preserve_series=True)

    if lag_max >= y.shape[0]:
        raise ValueError(
            f"lag_max ({lag_max}) must be < length of the "
            f"series ({y.shape[0]})"
        )

    # ax0 is simply the series itself
    x0 = np.arange(y.shape[0])
    xlabs = None
    if hasattr(y, 'index'):
        xlabs = y.index.tolist()
        y = y.values
    series_kwargs = {} if not series_kwargs else series_kwargs
    ax0.plot(x0, y, **series_kwargs)
    if title:
        ax0.set_title(title)

    if xlabs is not None:
        # TODO: eventually would be nice to do some well spaced xticks
        pass

    # ax1 is the ACF, so we can just use our ACF plotting func
    acf_kwargs = {} if not acf_kwargs else acf_kwargs
    plot_acf(y, ax=ax1, show=False, title='ACF', lags=lag_max, **acf_kwargs)

    # ax2 is simply the histogram
    hist_kwargs = {} if not hist_kwargs else hist_kwargs
    _ = ax2.hist(y, bins=bins, **hist_kwargs)  # noqa
    ax2.set_title("Frequency")

    fig.tight_layout()
    return _show_or_return(fig, show)



# -*- coding: utf-8 -*-

from pmdarima.datasets import load_heartrate, load_lynx, load_wineind,\
    load_woolyrnq, load_ausbeer, load_austres, load_gasoline, \
    load_airpassengers, load_taylor, load_msft, load_sunspots, _base as base

import numpy as np
import pandas as pd
import os
import shutil

from numpy.testing import assert_array_equal
import pytest


def _inner_load(f):
    n = None
    for as_series in (True, False):
        x = f(as_series=as_series)

        # ensure shape is same for both
        if n is None:
            n = x.shape[0]
        else:
            assert x.shape[0] == n

        if as_series:
            assert isinstance(x, pd.Series)
        else:
            assert isinstance(x, np.ndarray)


# Simply test loading the datasets and that we get the expected type
@pytest.mark.parametrize(
    'f', [load_heartrate,
          load_lynx,
          load_wineind,
          load_woolyrnq,
          load_ausbeer,
          load_austres,
          load_taylor,
          load_airpassengers])
def test_load(f):
    _inner_load(f)


@pytest.mark.parametrize(
    'f', [load_msft])
def test_df_loads(f):
    df = f()
    assert isinstance(df, pd.DataFrame)


@pytest.mark.parametrize(
    'f, cache_name', [
        pytest.param(load_sunspots, 'sunspots'),
    ])
def test_load_from_gzip(f, cache_name):
    _inner_load(f)
    assert cache_name in base._cache


@pytest.mark.parametrize(
    'func, key', [
        pytest.param(load_gasoline, 'gasoline'),
    ]
)
def test_load_from_web(func, key):
    # make sure there is no data folder
    disk_cache_folder = base.get_data_cache_path()
    if os.path.exists(disk_cache_folder):
        shutil.rmtree(disk_cache_folder)

    try:
        # loads from web
        y = func(as_series=False)

        # show the key is in _cache
        assert key in base._cache

        # show exists on disk
        assert os.path.exists(os.path.join(disk_cache_folder, key + '.csv.gz'))

        # pop from cache so we can load it from disk
        base._cache.pop(key)
        x = func(as_series=True)  # true for coverage

        assert_array_equal(y, x.values)

    finally:
        if os.path.exists(disk_cache_folder):
            shutil.rmtree(disk_cache_folder)



# -*- coding: utf-8 -*-
#
# Handle inconsistencies in the statsmodels API versions

from collections.abc import Iterable
from packaging.version import Version
import statsmodels as sm

__all__ = [
    'bind_df_model'
]

_sm_version = sm.__version__


def bind_df_model(model_fit, arima_results):
    """Set model degrees of freedom.

    Older versions of statsmodels don't handle this issue. Sets the
    model degrees of freedom in place if not already present.

    Parameters
    ----------
    model_fit : ARMA, ARIMA or SARIMAX
        The fitted model.

    arima_results : ModelResultsWrapper
        The results wrapper.
    """
    if not hasattr(arima_results, 'df_model'):
        df_model = model_fit.k_exog + model_fit.k_trend + \
            model_fit.k_ar + model_fit.k_ma + \
            model_fit.k_seasonal_ar + model_fit.k_seasonal_ma
        setattr(arima_results, 'df_model', df_model)


def check_seasonal_order(order):
    """Check the seasonal order

    Statsmodels 0.11.0 introduced a check for seasonal order == 1 that can
    raise a ValueError, but some of our old defaults allow for m == 1 in an
    otherwise null seasonal order.

    Parameters
    ----------
    order : tuple
        The existing seasonal order
    """

    # If order[0] is an iterable, but not a string then we don't perform check.
    # Otherwise we perform the check and override order if it satisfies check.
    # See issue#370: https://github.com/alkaline-ml/pmdarima/issues/370
    if isinstance(order[0], Iterable) and not isinstance(order[0], str):
        return order
    else:
        if sum(order[:3]) == 0 and order[-1] == 1:
            order = (0, 0, 0, 0)

        # user's order may be invalid, but we'll let statsmodels' validation
        # handle that.
        return order


def _use_sm13():
    return Version(sm.__version__) >= Version("0.13.0")



# -*- coding: utf-8 -*-

from .dates import *
from .fourier import *

# don't want to accidentally hoist `base` to top-level, since preprocessing has
# its own base
__all__ = [s for s in dir() if not (s.startswith("_") or s == 'base')]



# -*- coding: utf-8 -*-
#
# Author: Taylor Smith <taylor.smith@alkaline-ml.com>
#
# Tests for stationarity

from sklearn.base import BaseEstimator
from sklearn.linear_model import LinearRegression

from statsmodels import api as sm
from abc import ABCMeta, abstractmethod
import numpy as np

from ..compat.numpy import DTYPE
from ..decorators import deprecated
from ..utils.array import c, diff, check_endog
from .approx import approx

# since the C import relies on the C code having been built with Cython,
# and since the platform might name the .so file something funky (like
# _arima.cpython-35m-darwin.so), import this absolutely and not relatively.
from ._arima import C_tseries_pp_sum

__all__ = [
    'ADFTest',
    'KPSSTest',
    'PPTest'
]


class _BaseStationarityTest(BaseEstimator, metaclass=ABCMeta):
    @staticmethod
    def _base_case(x):
        # if x is empty, return false so the other methods return False
        if (x is None) or (x.shape[0] == 0):
            return False
        return True

    @staticmethod
    def _embed(x, k):
        # lag the vector and put the lags into columns
        n = x.shape[0]
        if k > n:
            raise ValueError("k cannot exceed y dim")

        rows = [
            # so, if k=2, it'll be (x[1:n], x[:n-1])
            x[j:n - i] for i, j in enumerate(range(k - 1, -1, -1))
        ]
        return np.asarray(rows)
        # return np.array([x[1:], x[:m]])


class _DifferencingStationarityTest(_BaseStationarityTest, metaclass=ABCMeta):
    """Provides the base class for stationarity tests such as the
    Kwiatkowski–Phillips–Schmidt–Shin, Augmented Dickey-Fuller and the
    Phillips–Perron tests. These tests are used to determine whether a time
    series is stationary.
    """
    def __init__(self, alpha):
        self.alpha = alpha

    @abstractmethod
    def should_diff(self, x):
        """Test whether the time series is stationary or it needs differencing.

        Parameters
        ----------
        x : array-like, shape=(n_samples,)
            The time series vector.

        Returns
        -------
        pval : float
            The computed P-value of the test.

        sig : bool
            Whether the P-value is significant at the ``alpha`` level.
            More directly, whether to difference the time series.
        """

    # TODO: REMOVE ME in v1.4.0
    @deprecated(use_instead="should_diff")
    def is_stationary(self, x):
        """Test whether the time series is stationary.

        Parameters
        ----------
        x : array-like, shape=(n_samples,)
            The time series vector.

        Returns
        -------
        pval : float
            The computed P-value of the test.

        sig : bool
            Whether the P-value is significant at the ``alpha`` level.
            More directly, whether to difference the time series.
        """
        return self.should_diff(x)


class KPSSTest(_DifferencingStationarityTest):
    """Conduct a KPSS test for stationarity.

    In econometrics, Kwiatkowski–Phillips–Schmidt–Shin (KPSS) tests are used
    for testing a null hypothesis that an observable time series is stationary
    around a deterministic trend (i.e. trend-stationary) against the
    alternative of a unit root.

    Parameters
    ----------
    alpha : float, optional (default=0.05)
        Level of the test

    null : str, optional (default='level')
        Whether to fit the linear model on the one vector, or an arange.
        If ``null`` is 'trend', a linear model is fit on an arange, if
        'level', it is fit on the one vector.

    lshort : bool, optional (default=True)
        Whether or not to truncate the ``l`` value in the C code.

    Notes
    -----
    This test is generally used indirectly via the
    :func:`pmdarima.arima.ndiffs` function, which computes the
    differencing term, ``d``.

    References
    ----------
    .. [1] R's tseries KPSS test source code: http://bit.ly/2eJP1IU
    """
    _valid = {'trend', 'null'}
    tablep = c(0.01, 0.025, 0.05, 0.10)

    def __init__(self, alpha=0.05, null='level', lshort=True):
        super(KPSSTest, self).__init__(alpha=alpha)

        self.null = null
        self.lshort = lshort

    def should_diff(self, x):
        """Test whether the time series is stationary or needs differencing.

        Parameters
        ----------
        x : array-like, shape=(n_samples,)
            The time series vector.

        Returns
        -------
        pval : float
            The computed P-value of the test.

        sig : bool
            Whether the P-value is significant at the ``alpha`` level.
            More directly, whether to difference the time series.
        """
        if not self._base_case(x):
            return np.nan, False

        # ensure vector
        x = check_endog(x, dtype=DTYPE, preserve_series=False)
        n = x.shape[0]

        # check on status of null
        null = self.null

        # fit a model on an arange to determine the residuals
        if null == 'trend':
            t = np.arange(n).reshape(n, 1)

            # these numbers came out of the R code.. I've found 0 doc for these
            table = c(0.216, 0.176, 0.146, 0.119)
        elif null == 'level':
            t = np.ones(n).reshape(n, 1)

            # these numbers came out of the R code.. I've found 0 doc for these
            table = c(0.739, 0.574, 0.463, 0.347)
        else:
            raise ValueError("null must be one of %r" % self._valid)

        # fit the model
        lm = LinearRegression().fit(t, x)
        e = x - lm.predict(t)  # residuals

        s = np.cumsum(e)
        eta = (s * s).sum() / (n**2)
        s2 = (e * e).sum() / n

        # scalar, denom = 10, 14
        # if self.lshort:
        #     scalar, denom = 3, 13
        # l_ = int(np.trunc(scalar * np.sqrt(n) / denom))
        if self.lshort:
            l_ = int(np.trunc(4 * (n / 100) ** 0.25))
        else:
            l_ = int(np.trunc(12 * (n / 100) ** 0.25))

        # compute the C subroutine
        s2 = C_tseries_pp_sum(e, n, l_, s2)
        stat = eta / s2

        # do approximation
        _, pval = approx(table, self.tablep, xout=stat, rule=2)

        # R does a test for rule=1, but we don't want to do that, because they
        # just do it to issue a warning in case the P-value is smaller/greater
        # than the printed value is.
        return pval[0], pval[0] < self.alpha


class ADFTest(_DifferencingStationarityTest):
    """Conduct an ADF test for stationarity.

    In statistics and econometrics, an augmented Dickey–Fuller test (ADF)
    tests the null hypothesis of a unit root is present in a time series
    sample. The alternative hypothesis is different depending on which version
    of the test is used, but is usually stationarity or trend-stationarity. It
    is an augmented version of the Dickey–Fuller test for a larger and more
    complicated set of time series models.

    Parameters
    ----------
    alpha : float, optional (default=0.05)
        Level of the test

    k : int, optional (default=None)
        The drift parameter. If ``k`` is None, it will be set to:
        ``np.trunc(np.power(x.shape[0] - 1, 1 / 3.0))``

    Notes
    -----
    This test is generally used indirectly via the
    :func:`pmdarima.arima.ndiffs` function, which computes the
    differencing term, ``d``.

    ADF test does not perform as close to the R code as do the KPSS and PP
    tests. This is due to the fact that is has to use statsmodels OLS
    regression for std err estimates rather than the more robust sklearn
    LinearRegression.

    References
    ----------
    .. [1] https://wikipedia.org/wiki/Augmented_Dickey–Fuller_test
    .. [2] R's tseries ADF source code: https://bit.ly/2EnvM5V
    """
    table = np.array([
        (-4.38, -3.95, -3.60, -3.24, -1.14, -0.80, -0.50, -0.15),
        (-4.15, -3.80, -3.50, -3.18, -1.19, -0.87, -0.58, -0.24),
        (-4.04, -3.73, -3.45, -3.15, -1.22, -0.90, -0.62, -0.28),
        (-3.99, -3.69, -3.43, -3.13, -1.23, -0.92, -0.64, -0.31),
        (-3.98, -3.68, -3.42, -3.13, -1.24, -0.93, -0.65, -0.32),
        (-3.96, -3.66, -3.41, -3.12, -1.25, -0.94, -0.66, -0.33)
    ])

    tablen = table.shape[1]
    tableT = c(25, 50, 100, 250, 500, 100000)
    tablep = c(0.01, 0.025, 0.05, 0.10, 0.90, 0.95, 0.975, 0.99)

    def __init__(self, alpha=0.05, k=None):
        super(ADFTest, self).__init__(alpha=alpha)

        self.k = k
        if k is not None and k < 0:
            raise ValueError('k must be a positive integer (>= 0)')

    @staticmethod
    def _ols(x, y, z, k):
        n = y.shape[0]
        yt = z[:, 0]  # type: np.ndarray
        tt = np.arange(k - 1, n)

        # R does [k:n].. but that's 1-based indexing and inclusive on the tail
        xt1 = x[tt]

        # make tt inclusive again (it was used as a mask before)
        tt += 1

        # the array that will create the LM:
        _n = xt1.shape[0]  # row dim for predictors
        X = np.hstack([np.ones(_n).reshape((_n, 1)),
                       xt1.reshape((_n, 1)),
                       tt.reshape((_n, 1))])

        if k > 1:
            yt1 = z[:, 1:k]  # R had 2:k
            X = np.hstack([X, yt1])

        # fit the linear regression - this one is a bit strange in that we
        # are using OLS from statsmodels rather than LR from sklearn. This is
        # because we need the std errors, and sklearn does not have a way to
        # store them.
        return sm.OLS(yt, X, hasconst=True).fit(method='qr')

    @staticmethod
    def _ols_std_error(res):
        stderrs = res.bse  # this was a pain in the ARSE to locate
        return res.params[1] / stderrs[1]

    def should_diff(self, x):
        """Test whether the time series is stationary or needs differencing.

        Parameters
        ----------
        x : array-like, shape=(n_samples,)
            The time series vector.

        Returns
        -------
        pval : float
            The computed P-value of the test.

        sig : bool
            Whether the P-value is significant at the ``alpha`` level.
            More directly, whether to difference the time series.
        """
        if not self._base_case(x):
            return np.nan, False

        # ensure vector
        x = check_endog(x, dtype=DTYPE, preserve_series=False)

        # if k is none...
        k = self.k
        if k is None:
            k = np.trunc(np.power(x.shape[0] - 1, 1 / 3.0))

        # See [2] for the R source. This is L153 - L160
        k = int(k) + 1
        y = diff(x)  # diff(as.vector(x, mode='double'))
        n = y.shape[0]
        z = self._embed(y, k).T  # Same as R embed(x, k)

        # Compute ordinary least squares
        res = self._ols(x, y, z, k)
        STAT = self._ols_std_error(res)

        # In the past we assigned to the np memory view which is slower
        tableipl = np.array([
            approx(self.tableT, self.table[:, i], xout=n, rule=2)[1]  # xt,yt
            for i in range(self.tablen)])

        # make sure to do 1 - x...
        _, interpol = approx(tableipl, self.tablep, xout=STAT, rule=2)

        # Added in v1.1.0. Not sure whether it's likely we'll hit it, but it's
        # how R is warning the user...
        # if np.isnan(approx(tableipl, self.tablep, xout=STAT, rule=1)[1]):
        #     if interpol == self.tablep.min():
        #         warnings.warn("p-value is smaller than printed value")
        #     else:
        #         warnings.warn("p-value is larger than printed value")

        # pval = 1 - interpol[0]  # explosive
        pval = interpol[0]  # stationarity

        # in the R code, here is where the P value warning is tested again...
        # else if (test == "adf")
        #     suppressWarnings(dodiff < - tseries::adf.test(x)$p.value > alpha)
        return pval, pval > self.alpha  # > since not 1-pval


class PPTest(_DifferencingStationarityTest):
    """Conduct a PP test for stationarity.

    In statistics, the Phillips–Perron test (named after Peter C. B.
    Phillips and Pierre Perron) is a unit root test. It is used in time series
    analysis to test the null hypothesis that a time series is integrated of
    order 1. It builds on the Dickey–Fuller test of the null hypothesis
    ``p = 0``.

    Parameters
    ----------
    alpha : float, optional (default=0.05)
        Level of the test

    lshort : bool, optional (default=True)
        Whether or not to truncate the ``l`` value in the C code.

    Notes
    -----
    This test is generally used indirectly via the
    :func:`pmdarima.arima.ndiffs` function, which computes the
    differencing term, ``d``.

    The R code allows for two types of tests: 'Z(alpha)' and 'Z(t_alpha)'.
    Since sklearn does not allow extraction of std errors from the linear
    model fit, ``t_alpha`` is much more difficult to achieve, so we do not
    allow that variant.

    References
    ----------
    .. [1] R's tseries PP test source code: http://bit.ly/2wbzx6V
    """
    table = -np.array([
        (22.5, 25.7, 27.4, 28.4, 28.9, 29.5),
        (19.9, 22.4, 23.6, 24.4, 24.8, 25.1),
        (17.9, 19.8, 20.7, 21.3, 21.5, 21.8),
        (15.6, 16.8, 17.5, 18.0, 18.1, 18.3),
        (3.66, 3.71, 3.74, 3.75, 3.76, 3.77),
        (2.51, 2.60, 2.62, 2.64, 2.65, 2.66),
        (1.53, 1.66, 1.73, 1.78, 1.78, 1.79),
        (0.43, 0.65, 0.75, 0.82, 0.84, 0.87)
    ]).T

    tablen = table.shape[1]
    tableT = c(25, 50, 100, 250, 500, 100000).astype(DTYPE)
    tablep = c(0.01, 0.025, 0.05, 0.10, 0.90, 0.95, 0.975, 0.99)

    def __init__(self, alpha=0.05, lshort=True):
        super(PPTest, self).__init__(alpha=alpha)

        self.lshort = lshort

    def should_diff(self, x):
        """Test whether the time series is stationary or needs differencing.

        Parameters
        ----------
        x : array-like, shape=(n_samples,)
            The time series vector.

        Returns
        -------
        pval : float
            The computed P-value of the test.

        sig : bool
            Whether the P-value is significant at the ``alpha`` level.
            More directly, whether to difference the time series.
        """
        if not self._base_case(x):
            return np.nan, False

        # ensure vector
        x = check_endog(x, dtype=DTYPE, preserve_series=False)

        # embed the vector. This is some funkiness that goes on in the R
        # code... basically, make a matrix where the column (rows if not T)
        # are lagged windows of x
        z = self._embed(x, 2)  # Same as R t(embed(x, k))
        yt = z[0, :]
        yt1 = z[1, :]  # type: np.ndarray

        # fit a linear model to a predictor matrix
        n = yt.shape[0]
        tt = (np.arange(n) + 1) - (n / 2.0)
        X = np.array([np.ones(n), tt, yt1]).T
        res = LinearRegression().fit(X, yt)  # lm(yt ~ 1 + tt + yt1)
        coef = res.coef_

        # check for singularities - do we want to do this??? in the R code,
        # it happens. but the very same lm in the R code is rank 3, and here
        # it is rank 2. Should we just ignore?...
        # if res.rank_ < 3:
        #     raise ValueError('singularities in regression')

        u = yt - res.predict(X)  # residuals
        ssqru = (u * u).sum() / float(n)

        scalar = 12 if not self.lshort else 4
        l_ = int(np.trunc(scalar * np.power(n / 100.0, 0.25)))
        ssqrtl = C_tseries_pp_sum(u, n, l_, ssqru)

        # define trm vals
        n2 = n * n
        syt11n = (yt1 * (np.arange(n) + 1)).sum()  # sum(yt1*(1:n))
        trm1 = n2 * (n2 - 1) * (yt1 ** 2).sum() / 12.0

        # R code: # n*sum(yt1*(1:n))^2
        trm2 = n * (syt11n ** 2)

        # R code: n*(n+1)*sum(yt1*(1:n))*sum(yt1)
        trm3 = n * (n + 1) * syt11n * yt1.sum()
        trm4 = (n * (n + 1) * (2 * n + 1) * (yt1.sum() ** 2)) / 6.0
        dx = trm1 - trm2 + trm3 - trm4

        # if self.typ == 'alpha':
        alpha = coef[2]  # it's the last col...
        STAT = n * (alpha - 1) - (n ** 6) / (24.0 * dx) * (ssqrtl - ssqru)

        tableipl = np.array([
            approx(self.tableT, self.table[:, i], xout=n, rule=2)[1]
            for i in range(self.tablen)])

        # we don't do 1 - pval, so check for GREATER THAN
        _, interpol = approx(tableipl, self.tablep, xout=STAT, rule=2)
        pval = interpol[0]

        # in the R code, here is where the P value warning is tested again...
        return pval, pval > self.alpha



# -*- coding: utf-8 -*-
#
# Author: Krishna Sunkara (kpsunkara)
#
# Re-entrant, reusable context manager to store execution context. Introduced
# in pmdarima 1.5.0 (see #221), redesigned not to use thread locals in #273
# (see #275 for context).

from abc import ABC, abstractmethod
from enum import Enum
import collections

__all__ = ['AbstractContext', 'ContextStore', 'ContextType']


class _CtxSingleton:
    """Singleton class to store context information"""
    store = {}


_ctx = _CtxSingleton()


class ContextType(Enum):
    """Context Type Enumeration

    An enumeration of Context Types known to :class:`ContextStore`
    """
    EMPTY = 0
    STEPWISE = 1


class AbstractContext(ABC):
    """An abstract context manager to store execution context.

    A generic, re-entrant, reusable context manager to store
    execution context. Has helper methods to iterate over the context info
    and provide a string representation of the context info.
    """
    def __init__(self, **kwargs):
        # remove None valued entries,
        # since __getattr__ returns None if an attr is not present
        self.props = {k: v for k, v in kwargs.items() if v is not None} \
            if kwargs else {}

    def __enter__(self):
        ContextStore._add_context(self)

    def __exit__(self, exc_type, exc_val, exc_tb):
        ContextStore._remove_context(self)

    def __getattr__(self, item):
        """Lets us access, e.g., ``ctx.max_steps`` even if not in a context"""
        return self.props[item] if item in self.props else None

    def __contains__(self, item):
        return item in self.props

    def __getitem__(self, item):
        return self.props[item] if item in self.props else None

    def __iter__(self):
        return iter(self.props)

    def keys(self):
        return self.props.keys()

    def values(self):
        return self.props.values()

    def items(self):
        return self.props.items()

    def update(self, other):
        parent_props = dict(other)
        parent_props.update(self.props)
        self.props = parent_props

    def __repr__(self):
        return self.props.__repr__()

    @abstractmethod
    def get_type(self):
        """Get the ContextType"""


class _emptyContext(AbstractContext):
    """An empty context for convenience use"""

    def __init__(self):
        super(_emptyContext, self).__init__()

    def get_type(self):
        """Indicates we are not in a context manager"""
        return ContextType.EMPTY


class ContextStore:
    """A class to wrap access to the global context store

    This class hosts static methods to wrap access to and encapsulate the
    singleton content store instance
    """
    @staticmethod
    def get_context(context_type):
        """Returns most recently added instance of given Context Type

        Parameters
        ----------
        context_type : ContextType
            Context type to retrieve from the store

        Returns
        -------
        res : AbstractContext
            An instance of AbstractContext subclass or None
        """
        if not isinstance(context_type, ContextType):
            raise ValueError('context_type must be an instance of ContextType')

        if context_type in _ctx.store and len(_ctx.store[context_type]) > 0:
            return _ctx.store[context_type][-1]

        # If not present
        return None

    @staticmethod
    def get_or_default(context_type, default):
        """Returns most recent instance of given Context Type or default

        Parameters
        ----------
        context_type : ContextType
            Context type to retrieve from the store

        default : AbstractContext
            Value to return in case given context does not exist

        Returns
        -------
        ctx : AbstractContext
            An instance of AbstractContext subclass or default
        """
        ctx = ContextStore.get_context(context_type)
        return ctx if ctx else default

    @staticmethod
    def get_or_empty(context_type):
        """Returns recent instance of given Context Type or an empty context

        Parameters
        ----------
        context_type : ContextType
            Context type to retrieve from the store

        Returns
        -------
        res : AbstractContext
            An instance of AbstractContext subclass
        """
        return ContextStore.get_or_default(context_type, _emptyContext())

    @staticmethod
    def _add_context(ctx):
        """Add given instance of AbstractContext subclass to context store

        This private member is only called by ``AbstractContext.__init__()``

        if the given ctx is nested, merge parent context, to support
        following usage:

        Examples
        --------
        >>> from pmdarima.arima import StepwiseContext, auto_arima
        >>> with StepwiseContext(max_steps=10):
        ...     with StepwiseContext(max_dur=30):
        ...         auto_arima(samp,...)

        This is identical to:
        >>> from contextlib import ExitStack
        ... stack = ExitStack()
        ... outer_ctx = StepwiseContext(max_steps=10)
        ... inner_ctx = StepwiseContext(max_dur=30)
        ... stack.enter_context(outer_ctx)
        ... stack.enter_context(inner_ctx)
        ... with stack:
        ...     auto_arima(samp, ...)


        However, the nested context can override parent context. In the
        example below, the effective context for inner most call to
        ``auto_arima(...)`` is: ``max_steps=15, max_dur=30``. The effective
        context for the second call to ``auto_arima(..)`` is: ``max_steps=10``

        >>> with StepwiseContext(max_steps=10):
        ...     with StepwiseContext(max_steps=15, max_dur=30):
        ...         auto_arima(samp,...)
        ...
        ...     auto_arima(samp,...)
        """
        if not isinstance(ctx, AbstractContext):
            raise ValueError('ctx must be be an instance of AbstractContext')

        # if given Context Type is not present into store, make an entry
        context_type = ctx.get_type()
        if context_type not in _ctx.store:
            _ctx.store[context_type] = collections.deque()

        # if the context is nested, merge with parent's context
        if len(_ctx.store[context_type]) > 0:
            parent = _ctx.store[context_type][-1]
            ctx.update(parent)

        _ctx.store[context_type].append(ctx)

    @staticmethod
    def _remove_context(ctx):
        """Removes the most recently added context of given Context Type

        This private member is only used by ``AbstractContext``
        :param ctx:
        :return: None
        """
        if not isinstance(ctx, AbstractContext):
            raise ValueError('ctx must be be an instance of AbstractContext')

        context_type = ctx.get_type()

        if context_type not in _ctx.store or \
                len(_ctx.store[context_type]) == 0:
            return

        _ctx.store[context_type].pop()



# Test the approximation function

from pmdarima.arima.approx import approx, _regularize
from pmdarima.utils.array import c
from pmdarima.arima.stationarity import ADFTest

from numpy.testing import assert_array_almost_equal
import numpy as np

import pytest

table = c(0.216, 0.176, 0.146, 0.119)
tablep = c(0.01, 0.025, 0.05, 0.10)
stat = 1.01


def test_regularize():
    x, y = c(0.5, 0.5, 1.0, 1.5), c(1, 2, 3, 4)
    x, y = _regularize(x, y, 'mean')

    assert_array_almost_equal(x, np.array([0.5, 1.0, 1.5]))
    assert_array_almost_equal(y, np.array([1.5, 3.0, 4.0]))


def test_approx_rule1():
    # for rule = 1
    x, y = approx(table, tablep, stat, rule=1)
    assert_array_almost_equal(x, c(1.01))
    assert_array_almost_equal(y, c(np.nan))


def test_approx_rule2():
    # for rule = 2
    x, y = approx(table, tablep, stat, rule=2)
    assert_array_almost_equal(x, c(1.01))
    assert_array_almost_equal(y, c(0.01))


@pytest.mark.parametrize(
    'kwargs', [

        # fails for length differences
        dict(x=[1, 2, 3], y=[1, 2], xout=1.0),

        # fails for bad string
        dict(x=table, y=table, xout=1.0, method='bad-string'),

        # fails for bad length
        dict(x=[], y=[], xout=[], ties='mean'),

        # fails for bad length
        dict(x=[], y=[], xout=[], method='constant'),

        # fails for linear when < 2 samples
        dict(x=[1], y=[1], xout=[], method='linear', ties='ordered'),

        # fails for bad length
        dict(x=[], y=[], xout=[], method='constant'),

    ]
)
def test_corner_errors(kwargs):
    with pytest.raises(ValueError):
        approx(**kwargs)


def test_valid_corner():
    # *doesn't* fail for constant when < 2 samples
    approx(x=[1], y=[1], xout=[], method='constant', ties='ordered')


def test_approx_precision():
    # Test an example from R vs. Python to compare the expected values and
    # make sure we get as close as possible. This is from an ADFTest where k=1
    # and x=austres
    tableipl = np.array([[-4.0664],
                         [-3.7468],
                         [-3.462],
                         [-3.1572],
                         [-1.2128],
                         [-0.8928],
                         [-0.6104],
                         [-0.2704]])

    _, interpol = approx(tableipl, ADFTest.tablep, xout=-1.337233, rule=2)
    assert np.allclose(interpol, 0.84880354)  # in R we get 0.8488036



# -*- coding: utf-8 -*-

from pmdarima.__check_build import raise_build_error

import pytest


def test_raise_build_error():
    try:
        # Raise a value error to pass into the raise_build_error
        # to assert it turns it into an ImportError
        raise ValueError("this is a dummy err msg")
    except ValueError as v:
        with pytest.raises(ImportError):
            raise_build_error(v)



# -*- coding: utf-8 -*-
#
# Taylor G Smith <taylor.smith@alkaline.ml>
#
# Wrapped functions
from functools import wraps
import warnings

from statsmodels.tsa.stattools import acf as sm_acf, pacf as sm_pacf

__all__ = [
    'acf',
    'pacf'
]

# TODO: remove all explicit args/kwargs, making them *args, **kwargs


def inheritdoc(parent):
    """Inherit documentation from a parent

    Parameters
    ----------
    parent : callable
        The parent function or class that contains the sought-after
        docstring. If it doesn't have a docstring, this might behave
        in unexpected ways.
        
    Examples
    --------
    >>> def a(x=1):
    ...     '''This is documentation'''
    ...     return x
    ...
    >>> @inheritdoc(a)
    ... def b(x):
    ...     return 2 * a(x)
    ...
    >>> print(b.__doc__)
    This is documentation

    >>> print(b(2))
    4
    """
    def wrapper(func):
        # Assign the parent docstring to the child
        func.__doc__ = parent.__doc__

        @wraps(func)
        def caller(*args, **kwargs):
            return func(*args, **kwargs)
        return caller
    return wrapper


@inheritdoc(parent=sm_acf)
def acf(
    x,
    nlags=None,
    qstat=False,
    fft=None,
    alpha=None,
    missing='none',
    adjusted=False,
):
    return sm_acf(
        x=x,
        nlags=nlags,
        qstat=qstat,
        fft=fft,
        alpha=alpha,
        missing=missing,
        adjusted=adjusted,
    )


@inheritdoc(parent=sm_pacf)
def pacf(x, nlags=None, method='ywadjusted', alpha=None):
    # Handle kwarg deprecation in statsmodels 0.13.0
    if "unbiased" in method:
        warnings.warn(
            "The `*unbiased` methods have been deprecated in "
            "statsmodels >= 0.13.0. Please use `*adjusted` instead.",
            DeprecationWarning
        )
        method = method.replace("unbiased", "adjusted")
    elif method in ("ydu", "ywu", "ldu"):
        warnings.warn(
            "The `ydu`, `ywu`, and `ldu` methods have been deprecated in "
            "statsmodels >= 0.13.0. Please use `yda`, `ywa`, and `lda` "
            "instead.",
            DeprecationWarning
        )
        method = method.replace("u", "a")

    return sm_pacf(x=x, nlags=nlags, method=method, alpha=alpha)



# -*- coding: utf-8 -*-

__all__ = [
    'ModelFitWarning'
]


class ModelFitWarning(UserWarning):
    """Generic warning used for a model fit that might fail. More descriptive
    than simply trying to lump everything into a default UserWarning, which
    gives the user no insight into the reason for the warning apart from a
    (potentially) cryptic message. This allows the user to understand the
    warning emanates from an attempted model fit and originates from within
    the pmdarima package.
    """
    pass



# -*- coding: utf-8 -*-
#
# Author: Charles Drotar <drotarcharles@gmail.com>
#
# Patch backend for sklearn
from packaging.version import Version

import sklearn
from sklearn.exceptions import NotFittedError

__all__ = [
    'check_is_fitted',
    'if_delegate_has_method',
    'safe_indexing',
]


def check_is_fitted(estimator, attributes):
    """Ensure the model has been fitted

    Typically called at the beginning of an operation on a model that requires
    having been fit. Raises a ``NotFittedError`` if the model has not been
    fit.

    This is an adaptation of scikit-learn's ``check_is_fitted``, which has been
    changed recently in a way that is no longer compatible with our package.

    Parameters
    ----------
    estimator : estimator instance,
        The estimator that will be checked to see if it is fitted.

    attributes : str or iterable
        The attributes to check for
    """
    if isinstance(attributes, str):
        attributes = [attributes]
    if not hasattr(attributes, "__iter__"):
        raise TypeError("attributes must be a string or iterable")
    for attr in attributes:
        if hasattr(estimator, attr):
            return
    raise NotFittedError("Model has not been fit!")


def safe_indexing(X, indices):
    """Slice an array or dataframe. This is deprecated in sklearn"""
    if hasattr(X, 'iloc'):
        return X.iloc[indices]
    # numpy:
    # TODO: this does not currently support axis 1
    if hasattr(X, 'ndim') and X.ndim == 2:
        return X[indices, :]
    # list or 1d array
    return X[indices]


def _estimator_has(attr):
    """Checks if the model has a given attribute.

    Meant to be used along with `sklearn.utils.metaestimators.available_if`

    Parameters
    ----------
    attr : str
        The attribute to check the calling object for

    Returns
    -------
    fn : callable
        A function that will either raise an `AttributeError` if the attribute
        does not exist, or True if it does.
    """
    def check(self):
        # raise original `AttributeError` if `attr` does not exist
        getattr(self, attr)
        return True

    return check


def if_delegate_has_method(attr):
    """Compat method to replace `sklearn.utils.metaestimators.if_delegate_has`

    Older versions (< 1.0.0) of sklearn support it, but newer versions use
    `available_if` instead.

    References
    ----------
    .. [1] https://git.io/JzKiv
    .. [2] https://git.io/JzKiJ
    """
    if Version(sklearn.__version__) < Version("1.0.0"):
        from sklearn.utils.metaestimators import if_delegate_has_method
        return if_delegate_has_method(attr)
    else:
        from sklearn.utils.metaestimators import available_if
        return available_if(_estimator_has(attr))
